==> 0.6.2.tar.gz: lib/src/model/field.dart <==
part of fields.model;

/// A reflected instance variable or getter of an [Object].
/// 
/// The type parameter `T` documents the expected type of the underlying 
/// object.  It is asserted to be assignable from the actual reflected type of
/// the underlying object in checked mode only.
class Field<T> {
    
  /// Creates a new field for the instance getter with the given [declaredName] 
  /// in the given [owner] object.
  /// 
  /// An optional [getter] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forGetter(
      Symbol declaredName,
      InstanceMirror owner,
     {MethodMirror getter,
      Field parent}) {
    if (getter == null) {
      getter = owner.type.declarations[declaredName];
      if (getter == null || 
          getter is! MethodMirror || 
          !getter.isGetter || 
          getter.isStatic) {
        throw new ArgumentError(
            '$declaredName is not an instance getter of'
            ' ${owner.type.qualifiedName}');
      }
    }
    final setter = new Symbol('${MirrorSystem.getName(declaredName)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[Ignore]) {
      return null;
    }   
    return new Field._(
        declaredName, 
        _typeOf(getter.returnType),
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        parent: parent,        
        editorClass: metadata[FieldEditorType],
        transform: metadata[UseValueTransform],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        name: metadata[Name],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.declarations.containsKey(setter) 
          || metadata[ReadOnly],
        isGetter: true);
  }
  
  /// Creates a new field for the instance variable with the given 
  /// [declaredName] in the given [owner] object.
  /// 
  /// An optional [variable] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forVariable(
      Symbol declaredName,
      InstanceMirror owner,
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.declarations[declaredName];
      if (variable == null || 
          variable is! VariableMirror || 
          variable.isStatic) {
        throw new ArgumentError(
            '$declaredName is not an instance variable of'
            ' ${owner.type.qualifiedName}');
      }
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[Ignore]) {
      return null;
    }
    return new Field._(
        declaredName, 
        _typeOf(variable.type),
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        parent: parent,        
        editorClass: metadata[FieldEditorType],
        transform: metadata[UseValueTransform],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        name: metadata[Name],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly],
        isVariable: true);
  }
  
  factory Field.synthetic(
      Symbol declaredName,
      T getter(),
     {setter(T),
      Field parent,
      Type editorType,
      ValueTransform transform,
      List<FieldAnnotation> annotations,
      String description,
      String name,
      bool isPrivate}) {
    return new Field._(
        declaredName, 
        T,
        getter,
        setter,
        parent: parent,        
        editorClass: (editorType == null) ? null : reflectClass(editorType),
        transform: transform,
        annotations: annotations,
        description: description,
        name: name,
        isPrivate: isPrivate,
        isReadOnly: setter == null);
  }
  
  /// The name of this field as declared in the [owner]'s class declaration.
  final Symbol declaredName;
  
  /// A description of this field or `null` if this field is not annotated with
  /// a [Description] annotation.
  final String description;
  
  /// The display name of this field.
  /// 
  /// By default this is equal to the [declaredName] of this field.  It may be 
  /// overriden by annotating the field with a [Name] annotation.
  final String name;
  
  /// The type of this field's [value].
  final Type type;
  
  /// Returns `true` if this field is declared as a getter.
  final bool isGetter;
  
  /// Returns `true` if this field is declared as private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  /// 
  /// A field is read-only if it reflects a `final` variable, a getter with no
  /// matching setter, or if it is annotated with [ReadOnly].
  final bool isReadOnly;
  
  bool get isSynthetic => !isGetter && !isVariable;
  
  /// Returns `true` if this field is declared as an instance variable.
  final bool isVariable;
  
  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
  
  /// A mirror on the class of this field's editor; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field editor for this field
  /// is typically based on the [type].  If the field is annotated with 
  /// [FieldEditorType] then this will be the reflected class of the annotated
  /// editor type.
  final ClassMirror editorClass;
  
  /// This field's preferred transform; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a transform for this field 
  /// is chosen by an editor based on the [type].  If the field is annotated 
  /// with [UseValueTransform] then this will be the annotated transform.
  final ValueTransform transform;
  
  /// Any field annotations that have been applied to this field outside of the
  /// set of field annotations defined in this library.
  /// 
  /// The set of field annotations defined in this library ([Name], 
  /// [ReadOnly], etc...) have a direct effect on one or more properties of
  /// this field and are therefore not included here.
  /// 
  /// This collection provides a mechanism for annotating fields with custom
  /// metadata that may be accessed later by custom editors.
  final Iterable<FieldAnnotation> annotations;
  
  FieldList _children;
  /// The children of this field.
  /// 
  /// A field's children does not include those whose [type] is determined to be 
  /// `simple`.  The definition of `simple` used here is more inclusive than the 
  /// term as used in `dart:mirrors` and includes `Map`, `List`, `dynamic`, 
  /// `Type`, and `Object`.
  FieldList get children => _children;
    
  Object _value;
  /// The current value of this field.
  Object 
    get value => _value;
    set value(Object value) {
      if(isReadOnly) {
        throw new StateError('Field $name is read-only.');
      }
      if (value != _value) {
        _setter(value);
        _value = value;
        _refreshChildren().then((_) {
          _values.add(value);
          if (parent != null) {
            parent._values.add(parent._value);
          }
        });
      }
    }
  
  final _values = new StreamController<Object>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<Object> get values => _values.stream;
  
  Function _getter;
  Function _setter;
  
  Field._(
      Symbol declaredName,
      this.type,
      this._getter,
      this._setter,
     {this.parent,      
      this.editorClass, 
      this.transform,
      this.annotations: const [],
      this.description,
      String name, 
      this.isPrivate: false, 
      this.isReadOnly: false,
      this.isGetter: false,
      this.isVariable: false})
      : this.declaredName = declaredName
      , this.name = (name == null) ? MirrorSystem.getName(declaredName) : name {
    assert(_isAssignableFrom(reflectType(T), reflectType(type)));
    _children = new FieldList(this);
    _value = _getter();
    if (!_isSimpleType(type)) {
      _children.refreshSource(_value);
    }
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = _getter();
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _values.add(value));
    }
  }
        
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => new Future(f.refreshValue));
  }
}

==> 0.6.4.tar.gz: lib/src/model/field.dart <==
part of fields.model;

/// A reflected instance variable or getter of an [Object].
/// 
/// The type parameter `T` documents the expected type of the underlying 
/// object.  It is asserted to be assignable from the actual reflected type of
/// the underlying object in checked mode only.
class Field<T> {
    
  /// Creates a new field for the instance getter with the given [declaredName] 
  /// in the given [owner] object.
  /// 
  /// An optional [getter] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forGetter(
      Symbol declaredName,
      InstanceMirror owner,
     {MethodMirror getter,
      Field parent}) {
    if (getter == null) {
      getter = owner.type.declarations[declaredName];
      if (getter == null || 
          getter is! MethodMirror || 
          !getter.isGetter || 
          getter.isStatic) {
        throw new ArgumentError(
            '$declaredName is not an instance getter of'
            ' ${owner.type.qualifiedName}');
      }
    }
    final setter = new Symbol('${MirrorSystem.getName(declaredName)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[Ignore]) {
      return null;
    }   
    return new Field<T>._(
        declaredName,
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        type: _typeOf(getter.returnType),
        parent: parent,        
        editorClass: metadata[FieldEditorType],
        transform: metadata[UseValueTransform],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        name: metadata[Name],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.declarations.containsKey(setter) 
          || metadata[ReadOnly],
        isGetter: true);
  }
  
  /// Creates a new field for the instance variable with the given 
  /// [declaredName] in the given [owner] object.
  /// 
  /// An optional [variable] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory Field.forVariable(
      Symbol declaredName,
      InstanceMirror owner,
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.declarations[declaredName];
      if (variable == null || 
          variable is! VariableMirror || 
          variable.isStatic) {
        throw new ArgumentError(
            '$declaredName is not an instance variable of'
            ' ${owner.type.qualifiedName}');
      }
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[Ignore]) {
      return null;
    }
    return new Field<T>._(
        declaredName,
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        type: _typeOf(variable.type),
        parent: parent,    
        editorClass: metadata[FieldEditorType],
        transform: metadata[UseValueTransform],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        name: metadata[Name],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly],
        isVariable: true);
  }
  
  factory Field.synthetic(
      Symbol declaredName,
      T getter(),
     {setter(T),
      Field parent,
      Type editorType,
      ValueTransform transform,
      List<FieldAnnotation> annotations,
      String description,
      String name,
      bool isPrivate}) {
    return new Field<T>._(
        declaredName,
        getter,
        setter,
        parent: parent,        
        editorClass: (editorType == null) ? null : reflectClass(editorType),
        transform: transform,
        annotations: annotations,
        description: description,
        name: name,
        isPrivate: isPrivate,
        isReadOnly: setter == null);
  }
  
  /// The name of this field as declared in the [owner]'s class declaration.
  final Symbol declaredName;
  
  /// A description of this field or `null` if this field is not annotated with
  /// a [Description] annotation.
  final String description;
  
  /// The display name of this field.
  /// 
  /// By default this is equal to the [declaredName] of this field.  It may be 
  /// overriden by annotating the field with a [Name] annotation.
  final String name;
  
  /// The type of this field's [value].
  final Type type;
  
  /// Returns `true` if this field is declared as a getter.
  final bool isGetter;
  
  /// Returns `true` if this field is declared as private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  /// 
  /// A field is read-only if it reflects a `final` variable, a getter with no
  /// matching setter, or if it is annotated with [ReadOnly].
  final bool isReadOnly;
  
  bool get isSynthetic => !isGetter && !isVariable;
  
  /// Returns `true` if this field is declared as an instance variable.
  final bool isVariable;
  
  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
  
  /// A mirror on the class of this field's editor; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a field editor for this field
  /// is typically based on the [type].  If the field is annotated with 
  /// [FieldEditorType] then this will be the reflected class of the annotated
  /// editor type.
  final ClassMirror editorClass;
  
  /// This field's preferred transform; this is `null` by default.
  /// 
  /// By default, this is `null` and the choice of a transform for this field 
  /// is chosen by an editor based on the [type].  If the field is annotated 
  /// with [UseValueTransform] then this will be the annotated transform.
  final ValueTransform transform;
  
  /// Any field annotations that have been applied to this field outside of the
  /// set of field annotations defined in this library.
  /// 
  /// The set of field annotations defined in this library ([Name], 
  /// [ReadOnly], etc...) have a direct effect on one or more properties of
  /// this field and are therefore not included here.
  /// 
  /// This collection provides a mechanism for annotating fields with custom
  /// metadata that may be accessed later by custom editors.
  final Iterable<FieldAnnotation> annotations;
  
  FieldList _children;
  /// The children of this field.
  /// 
  /// A field's children does not include those whose [type] is determined to be 
  /// `simple`.  The definition of `simple` used here is more inclusive than the 
  /// term as used in `dart:mirrors` and includes `Map`, `List`, `dynamic`, 
  /// `Type`, and `Object`.
  FieldList get children => _children;
    
  Object _value;
  /// The current value of this field.
  Object 
    get value => _value;
    set value(Object value) {
      if(isReadOnly) {
        throw new StateError('Field $name is read-only.');
      }
      if (value != _value) {
        _setter(value);
        _value = value;
        _refreshChildren().then((_) {
          _values.add(value);
          if (parent != null) {
            parent._values.add(parent._value);
          }
        });
      }
    }
  
  final _values = new StreamController<Object>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<Object> get values => _values.stream;
  
  Function _getter;
  Function _setter;
  
  Field._(
      Symbol declaredName,
      this._getter,
      this._setter,
     {Type type,
      this.parent,      
      this.editorClass, 
      this.transform,
      this.annotations: const [],
      this.description,
      String name, 
      this.isPrivate: false, 
      this.isReadOnly: false,
      this.isGetter: false,
      this.isVariable: false})
      : this.declaredName = declaredName
      , this.type = (type == null) ? T : type
      , this.name = (name == null) ? MirrorSystem.getName(declaredName) : name {
    assert(_isAssignableFrom(reflectType(T), reflectType(this.type)));
    _children = new FieldList(this);
    _value = _getter();
    if (!_isSimpleType(this.type)) {
      _children.refreshSource(_value);
    }
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = _getter();
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _values.add(value));
    }
  }
        
  Future _refreshChildren() {    
    if (children.isEmpty) {
      return new Future.value();
    }
    children.refreshSource(_value);
    return Future.forEach(children, (f) => new Future(f.refreshValue));
  }
}
